"""Module for deliveries of workflow files"""

import logging

from cg.apps import hk, lims
from cg.store import Store

LOG = logging.getLogger(__name__)


class DeliverAPI:
    """Deliver API for workflows files"""

    def __init__(
        self,
        db: Store,
        hk_api: hk.HousekeeperAPI,
        lims_api: lims.LimsAPI,
        case_tags: [str],
        sample_tags: [str],
    ):
        self.store = db
        self.hk_api = hk_api
        self.lims = lims_api
        self.case_tags = case_tags
        self.sample_tags = sample_tags

    def get_post_analysis_files(self, case: str, version, tags: list):

        if not version:
            last_version = self.hk_api.last_version(bundle=case)
        else:
            last_version = self.hk_api.version(bundle=case, date=version)

        return self.hk_api.files(bundle=case, version=last_version.id, tags=tags).all()

    def get_post_analysis_case_files(self, case: str, version, tags):
        """Link files from HK to cust inbox."""

        tagged_files = self.get_post_analysis_files(case, version, tags)
        case_obj = self.store.family_samples(case)
        sample_ids = [case_sample.sample.internal_id for case_sample in case_obj]

        case_files = []
        for file_obj in tagged_files:

            if any(tag.name in sample_ids for tag in file_obj.tags):
                continue

            if any(tag.name in self.case_tags for tag in file_obj.tags):
                case_files.append(file_obj)

        return case_files

    def get_post_analysis_sample_files(self, case: str, sample: str, version, tag):
        """Link files from HK to cust inbox."""

        all_files = self.get_post_analysis_files(case, version, tag)

        sample_files = []
        for file_obj in all_files:
            if not any(tag.name == sample for tag in file_obj.tags):
                continue

            if any(tag.name in self.sample_tags for tag in file_obj.tags):
                sample_files.append(file_obj)

        return sample_files

    def get_post_analysis_files_root_dir(self):
        return self.hk_api.get_root_dir()



"""Tests for the scout upload API"""
from pathlib import Path

import pytest
import yaml

from cg.meta.upload.scoutapi import UploadScoutAPI
from cg.store import Store

CASE_FILE_PATHS = ["multiqc"]

RESULT_KEYS = [
    "family",
    "human_genome_build",
    "rank_model_version",
    "sv_rank_model_version",
]

SAMPLE_FILE_PATHS = ["alignment_path", "chromograph", "vcf2cytosure"]


def test_unlinked_family_is_linked(upload_scout_api):
    """Test that is_family check fails when samples are not linked"""
    # GIVEN a upload scout api and case data for a family without linked individuals
    family_data = {
        "samples": [
            {"sample_id": "ADM2", "father": "0", "mother": "0"},
            {"sample_id": "ADM3", "father": "0", "mother": "0"},
        ]
    }
    # WHEN running the check if family is linked
    res = upload_scout_api._is_family_case(family_data)
    # THEN assert that the test returns False
    assert res is False


def test_family_is_linked(upload_scout_api):
    """Test that is_family returns true when samples are linked"""
    # GIVEN a upload scout api and case data for a linked family
    family_data = {
        "samples": [
            {"sample_id": "ADM1", "father": "ADM2", "mother": "ADM3"},
            {"sample_id": "ADM2", "father": "0", "mother": "0"},
            {"sample_id": "ADM3", "father": "0", "mother": "0"},
        ]
    }
    # WHEN running the check if family is linked
    res = upload_scout_api._is_family_case(family_data)
    # THEN assert that the test returns True
    assert res is True


@pytest.mark.parametrize("result_key", RESULT_KEYS)
def test_generate_config_adds_meta_result_key(
    result_key, analysis: Store.Analysis, upload_scout_api: UploadScoutAPI
):
    """Test that generate config adds rank model version"""
    # GIVEN a status db and hk with an analysis
    assert analysis  # WHEN generating the scout config for the analysis
    result_data = upload_scout_api.generate_config(
        analysis
    )  # THEN the config should contain the rank model version used
    assert result_data[result_key]


@pytest.mark.parametrize("file_path", SAMPLE_FILE_PATHS)
def test_generate_config_adds_sample_paths(
    file_path, analysis: Store.Analysis, upload_scout_api: UploadScoutAPI
):
    """Test that generate config adds vcf2cytosure file"""
    # GIVEN a status db and hk with an analysis
    assert analysis

    # WHEN generating the scout config for the analysis
    result_data = upload_scout_api.generate_config(analysis)

    # THEN the config should contain the sample file path for each sample
    for sample in result_data["samples"]:
        assert sample[file_path]


@pytest.mark.parametrize("file_path", CASE_FILE_PATHS)
def test_generate_config_adds_sample_paths(
    file_path, analysis: Store.Analysis, upload_scout_api: UploadScoutAPI
):
    """Test that generate config adds case file paths"""
    # GIVEN a status db and hk with an analysis
    assert analysis

    # WHEN generating the scout config for the analysis
    result_data = upload_scout_api.generate_config(analysis)

    # THEN the config should contain the case file path
    assert result_data[file_path]


def _file_exists_on_disk(file_path: Path):
    """Returns if the file exists on disk"""
    return file_path.exists()


def _file_is_yaml(file_path):
    """Returns if the file successfully was loaded as yaml"""
    data = None
    with open(file_path, "r") as stream:
        try:
            data = yaml.safe_load(stream)
        except yaml.YAMLError:
            return False
    return data is not None


def test_save_config_creates_file(upload_scout_api: UploadScoutAPI, tmp_file):
    """"Tests that save config creates a file"""

    # GIVEN a scout_config dict and a path to save it on
    scout_config = {"dummy": "data"}

    # WHEN calling method to save config
    upload_scout_api.save_config_file(upload_config=scout_config, file_path=tmp_file)

    # THEN the config should exist on disk
    assert _file_exists_on_disk(tmp_file)


def test_save_config_creates_yaml(upload_scout_api: UploadScoutAPI, tmp_file):
    """Tests that the file created by save_config_file create a yaml """

    # GIVEN a scout_config dict and a path to save it on
    scout_config = {"dummy": "data"}

    # WHEN calling method to save config
    upload_scout_api.save_config_file(upload_config=scout_config, file_path=tmp_file)

    # THEN the should be of yaml type
    assert _file_is_yaml(tmp_file)


def test_add_scout_config_to_hk(upload_scout_api: UploadScoutAPI, tmp_file):
    """Test that scout load config is added to housekeeper"""
    # GIVEN a hk_mock and a file path to scout load config
    tag_name = UploadScoutAPI.get_load_config_tag()
    housekeeper_api = upload_scout_api.housekeeper
    # GIVEN a hk mock that does not return a config file
    housekeeper_api.add_missing_tag(tag_name)
    # WHEN adding the file path to hk_api

    upload_scout_api.add_scout_config_to_hk(
        config_file_path=tmp_file, hk_api=housekeeper_api, case_id="dummy"
    )
    # THEN assert that the file path is added to hk
    # file added to hk-db
    assert housekeeper_api.is_file_added() is True
    assert housekeeper_api.is_file_included() is True


def test_add_scout_config_to_hk_existing_files(upload_scout_api: UploadScoutAPI, tmp_file):
    """Test that scout config is not updated in housekeeper if it already exists"""
    # GIVEN a hk_mock with an scout upload file and a file path to scout load config
    housekeeper_api = upload_scout_api.housekeeper
    # GIVEN that there are files in the hk mock
    assert housekeeper_api.files
    # WHEN adding the file path to hk_api
    with pytest.raises(FileExistsError):
        # THEN assert File exists exception is raised
        upload_scout_api.add_scout_config_to_hk(
            config_file_path=tmp_file, hk_api=housekeeper_api, case_id="dummy"
        )



"""Code for uploading observations data via CLI"""
import logging

import click

from cg.apps import loqus
from cg.exc import DuplicateRecordError, DuplicateSampleError
from cg.meta.upload.observations import UploadObservationsAPI

from .utils import LinkHelper

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--case_id", help="internal case id, leave empty to process all")
@click.option("-l", "--case-limit", type=int, help="maximum number of cases to upload")
@click.option("--dry-run", is_flag=True, help="only print cases to be processed")
@click.pass_context
def observations(context, case_id, case_limit, dry_run):
    """Upload observations from an analysis to LoqusDB."""

    click.echo(click.style("----------------- OBSERVATIONS ----------------"))

    loqus_apis = {
        "wgs": loqus.LoqusdbAPI(context.obj),
        "wes": loqus.LoqusdbAPI(context.obj, analysis_type="wes"),
    }
    status_api = context.obj["status"]
    hk_api = context.obj["housekeeper_api"]

    if case_id:
        families_to_upload = [status_api.family(case_id)]
    else:
        families_to_upload = status_api.observations_to_upload()

    nr_uploaded = 0
    for family_obj in families_to_upload:

        if case_limit is not None:
            if nr_uploaded >= case_limit:
                LOG.info("Uploaded %d cases, observations upload will now stop", nr_uploaded)
                return

        if not family_obj.customer.loqus_upload:
            LOG.info(
                "%s: %s not whitelisted for upload to loqusdb. Skipping!",
                family_obj.internal_id,
                family_obj.customer.internal_id,
            )
            continue

        if not LinkHelper.all_samples_data_analysis(family_obj.links, ["MIP", "", None]):
            LOG.info("%s: has non-MIP data_analysis. Skipping!", family_obj.internal_id)
            continue

        if not LinkHelper.all_samples_are_non_tumour(family_obj.links):
            LOG.info("%s: has tumour samples. Skipping!", family_obj.internal_id)
            continue

        analysis_list = LinkHelper.all_samples_list_analyses(family_obj.links)
        if not (len(set(analysis_list)) == 1 and analysis_list[0] in ("wes", "wgs")):
            LOG.info(
                "%s: Undetermined analysis type (wes or wgs) or mixed analyses. Skipping!",
                family_obj.internal_id,
            )
            continue

        analysis_type = analysis_list[0]

        if dry_run:
            LOG.info("%s: Would upload observations", family_obj.internal_id)
            continue

        api = UploadObservationsAPI(status_api, hk_api, loqus_apis[analysis_type])

        try:
            api.process(family_obj.analyses[0])
            LOG.info("%s: observations uploaded!", family_obj.internal_id)
            nr_uploaded += 1
        except (DuplicateRecordError, DuplicateSampleError) as error:
            LOG.info("%s: skipping observations upload: %s", family_obj.internal_id, error.message)
        except FileNotFoundError as error:
            LOG.info("%s: skipping observations upload: %s", family_obj.internal_id, error)



# -*- coding: utf-8 -*- import logging

import logging
import os
from pathlib import Path

import click
from cg.apps import hk, lims
from cg.meta.deliver import DeliverAPI
from cg.store import Store

LOG = logging.getLogger(__name__)

CASE_TAGS = []

SAMPLE_TAGS = ["bam", "bam-index", "cram", "cram-index"]


@click.group()
@click.pass_context
def deliver(context):
    """Deliver files to customer inbox."""
    context.obj["db"] = Store(context.obj["database"])
    context.obj["deliver_api"] = DeliverAPI(
        db=context.obj["db"],
        hk_api=hk.HousekeeperAPI(context.obj),
        lims_api=lims.LimsAPI(context.obj),
        case_tags=CASE_TAGS,
        sample_tags=SAMPLE_TAGS,
    )


@deliver.command()
@click.option("-V", "--version", help="version (date) of bundle")
@click.option("-t", "--tag", multiple=True, help="the housekeeper tag(s)")
@click.option(
    "-i",
    "--inbox",
    "inbox_path",
    default="/home/proj/production/customers/{customer}/inbox/{case}/",
    help="customer inbox",
)
@click.argument("case", required=False)
@click.pass_context
def inbox(context, case, version, tag, inbox_path):
    """Link files from Housekeeper to customer inbox."""

    if not case:
        _suggest_cases_to_deliver(context.obj["db"])
        context.abort()

    case_obj = context.obj["db"].family(case)
    if case_obj is None:
        LOG.error("Case '%s' not found.", case)
        context.abort()

    files = context.obj["deliver_api"].get_post_analysis_case_files(
        case=case, version=version, tags=tag
    )
    if not files:
        LOG.warning("No case files found")

    for file_obj in files:
        out_dir = Path(
            inbox_path.format(case=case_obj.name, customer=case_obj.customer.internal_id)
        )
        out_dir.mkdir(parents=True, exist_ok=True)

        out_path = _generate_case_delivery_path(case, case_obj, file_obj, out_dir)
        in_path = Path(file_obj.full_path)

        if not out_path.exists():
            os.link(in_path, out_path)
            LOG.info("linked file: %s -> %s", in_path, out_path)
        else:
            LOG.info("Target file exists: %s", out_path)

    link_obj = context.obj["db"].family_samples(case)
    if not link_obj:
        LOG.warning("No sample files found.")

    for case_sample in link_obj:
        sample_obj = case_sample.sample
        files = context.obj["deliver_api"].get_post_analysis_sample_files(
            case=case, sample=sample_obj.internal_id, version=version, tag=tag
        )

        if not files:
            LOG.warning("No sample files found for '%s'.", sample_obj.internal_id)

        for file_obj in files:
            out_dir = Path(
                inbox_path.format(case=case_obj.name, customer=case_obj.customer.internal_id)
            )
            out_dir = out_dir.joinpath(sample_obj.name)
            out_dir.mkdir(parents=True, exist_ok=True)

            out_path = _generate_sample_delivery_path(file_obj, out_dir, sample_obj)
            in_path = Path(file_obj.full_path)

            if not out_path.exists():
                os.link(in_path, out_path)
                LOG.info("linked file: %s -> %s", in_path, out_path)
            else:
                LOG.info("Target file exists: %s", out_path)


def _generate_case_delivery_path(case, case_obj, file_obj, out_dir):
    return Path(f"{out_dir / Path(file_obj.path).name.replace(case, case_obj.name)}")


def _generate_sample_delivery_path(file_obj, out_dir, sample_obj):
    return Path(
        f"{out_dir / Path(file_obj.path).name.replace(sample_obj.internal_id, sample_obj.name)}"
    )


def _suggest_cases_to_deliver(store):
    LOG.warning("provide a case, suggestions:")

    for case_obj in store.analyses_to_deliver(pipeline="balsamic", limit=50):
        click.echo(case_obj)



"""Test MIP get files and build bundle"""
import mock
import pytest

from snapshottest import Snapshot

import cg.meta.store.mip as store_mip
from cg.exc import (
    AnalysisNotFinishedError,
    BundleAlreadyAddedError,
)


@mock.patch("cg.meta.store.base.build_bundle")
@mock.patch("cg.meta.store.mip.parse_sampleinfo")
@mock.patch("cg.meta.store.mip.parse_config")
def test_add_mip_analysis_finished(
    mock_parse_config,
    mock_parse_sample,
    mock_build_bundle,
    config_stream,
    config_data_rna,
    sampleinfo_data,
    rna_deliverables_raw,
):
    """
    tests the function add_mip_analysis when passing a config file of a finished RNA analysis
    """
    # GIVEN a MIP RNA analysis configuration file
    mip_rna_config = config_stream["rna_config_store"]

    # WHEN gathering information from the MIP RNA analysis to store and the analysis is
    # finished
    mock_parse_config.return_value = config_data_rna
    mock_parse_sample.return_value = sampleinfo_data

    result = store_mip.add_mip_analysis(mip_rna_config)

    # THEN the function add_mip_analysis should call the function build_bundle with the correct
    # parameters and return a new bundle
    mock_build_bundle.assert_called()
    mock_build_bundle.assert_called_with(config_data_rna, sampleinfo_data, rna_deliverables_raw)
    assert result == mock_build_bundle(config_data_rna, sampleinfo_data, rna_deliverables_raw)


@mock.patch("cg.meta.store.mip.parse_sampleinfo")
@mock.patch("cg.meta.store.mip.parse_config")
def test_add_mip_analysis_not_finished(
    mock_parse_config, mock_parse_sample, config_stream, config_data_rna, sampleinfo_data
):
    """
    tests the function add_mip_analysis when passing a config file of an unfinished RNA analysis
    """
    # GIVEN a MIP RNA analysis configuration file

    mip_rna_config = config_stream["rna_config_store"]
    mock_parse_config.return_value = config_data_rna
    sampleinfo_data["is_finished"] = False
    mock_parse_sample.return_value = sampleinfo_data

    # WHEN  gathering information from the MIP RNA analysis to store and the analysis is not
    # finished

    with pytest.raises(AnalysisNotFinishedError) as exc_info:
        store_mip.add_mip_analysis(mip_rna_config)

    # THEN the correct exception should be raised
    assert exc_info.value.message == "analysis not finished"


def test_parse_config(snapshot: Snapshot, config_raw: dict):
    """
    tests the function parse_config against a snapshot
    """
    # GIVEN the raw MIP analysis config file

    # WHEN getting the files used to build the bundle
    mip_rna_parse_config = store_mip.parse_config(config_raw)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_rna_parse_config)


def test_parse_sampleinfo_data(snapshot: Snapshot, files_raw):
    """
    tests the function parse_sampleinfo_data against a snapshot
    """
    # GIVEN the raw MIP sample info file

    rna_sampleinfo = files_raw["rna_sampleinfo"]

    # WHEN getting the smaple info used to build the bundle
    mip_rna_parse_sampleinfo = store_mip.parse_sampleinfo(rna_sampleinfo)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_rna_parse_sampleinfo)



"""CLI function to compress FASTQ files into SPRING archives"""

import logging

import click

from cg.exc import CaseNotFoundError

from .helpers import (
    correct_spring_paths,
    get_fastq_cases,
    get_fastq_individuals,
    update_compress_api,
)

LOG = logging.getLogger(__name__)

# There is a list of problematic cases that we should skip
PROBLEMATIC_CASES = [
    "modernbee",
    "suremako",
    "expertalien",
    "wisestork",
    "richalien",
    "deepcub",
    "causalmite",
    "proudcollie",
    "loyalegret",
    "grandkoi",
    "fluenteagle",
    "lovingmayfly",
]

# List of cases used for validation that we should skip
VALIDATION_CASES = [
    "mintyeti",
    "topsrhino",
    "gladthrush",
    "cleanshrimp",
    "usablemarten",
    "casualgannet",
    "pumpedcat",
    "firstfawn",
    "helpedfilly",
    "daringpony",
    "strongbison",
    "proeagle",
    "easybeetle",
    "sharppigeon",
    "gamedeer",
    "keencalf",
    "epicasp",
    "safeguinea",
    "hotviper",
    "hotskink",
    "onemite",
    "busycolt",
    "rightmacaw",
    "intentcorgi",
    "vitalmouse",
    "lightprawn",
    "meetpossum",
    "strongman",
    "mintbaboon",
    "propercoral",
    "livingox",
    "keenviper",
    "sharpparrot",
    "moralgoat",
    "fleetjay",
    "bosssponge",
    "unitedbeagle",
]


@click.command("fastq")
@click.option("-c", "--case-id", type=str)
@click.option("-n", "--number-of-conversions", default=5, type=int, show_default=True)
@click.option("-t", "--ntasks", default=12, show_default=True, help="Number of tasks for slurm job")
@click.option("-m", "--mem", default=50, show_default=True, help="Memory for slurm job")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def fastq_cmd(context, case_id, number_of_conversions, ntasks, mem, dry_run):
    """ Find cases with FASTQ files and compress into SPRING """
    LOG.info("Running compress FASTQ")
    update_compress_api(context.obj["compress"], dry_run=dry_run, ntasks=ntasks, mem=mem)

    store = context.obj["db"]
    try:
        cases = get_fastq_cases(store, case_id)
    except CaseNotFoundError:
        return

    case_conversion_count = 0
    ind_conversion_count = 0
    for case in cases:
        # Keeps track on if all samples in a case have been converted
        case_converted = True
        if case_conversion_count >= number_of_conversions:
            break
        internal_id = case.internal_id
        if internal_id in [*PROBLEMATIC_CASES, *VALIDATION_CASES]:
            LOG.info("Skipping case %s", internal_id)
            continue

        LOG.info("Searching for FASTQ files in case %s", internal_id)
        for link_obj in case.links:
            sample_id = link_obj.sample.internal_id
            case_converted = context.obj["compress"].compress_fastq(sample_id)
            if case_converted is False:
                LOG.info("skipping individual %s", sample_id)
                continue
            ind_conversion_count += 1
        if case_converted:
            case_conversion_count += 1

    LOG.info(
        "%s Individuals in %s (completed) cases where compressed",
        ind_conversion_count,
        case_conversion_count,
    )


@click.command("fastq")
@click.option("-c", "--case-id")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def clean_fastq(context, case_id, dry_run):
    """Remove compressed FASTQ files, and update links in housekeeper to SPRING files"""
    LOG.info("Running compress clean FASTQ")
    compress_api = context.obj["compress"]
    update_compress_api(compress_api, dry_run=dry_run)

    store = context.obj["db"]
    samples = get_fastq_individuals(store, case_id)

    cleaned_inds = 0
    try:
        for sample_id in samples:
            res = compress_api.clean_fastq(sample_id)
            if res is False:
                LOG.info("skipping individual %s", sample_id)
                continue
            cleaned_inds += 1
    except CaseNotFoundError:
        return

    LOG.info("Cleaned fastqs in %s individuals", cleaned_inds)


@click.command("fix-spring")
@click.option("-b", "--bundle-name")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def fix_spring(context, bundle_name, dry_run):
    """Check if bundle(s) have non existing SPRING files and correct these"""
    LOG.info("Running compress clean FASTQ")
    compress_api = context.obj["compress"]
    update_compress_api(compress_api, dry_run=dry_run)
    hk_api = compress_api.hk_api
    correct_spring_paths(hk_api=hk_api, bundle_name=bundle_name, dry_run=dry_run)


@click.command("spring")
@click.argument("case-id")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_spring(context, case_id, dry_run):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""
    LOG.info("Running decompress spring")
    compress_api = context.obj["compress"]
    update_compress_api(compress_api, dry_run=dry_run)

    store = context.obj["db"]
    samples = get_fastq_individuals(store, case_id)

    decompressed_inds = 0
    try:
        for sample_id in samples:
            was_decompressed = compress_api.decompress_spring(sample_id)
            if was_decompressed is False:
                LOG.info("skipping individual %s", sample_id)
                continue
            decompressed_inds += 1
    except CaseNotFoundError:
        return

    LOG.info("Decompressed spring archives in %s individuals", decompressed_inds)



"""This script tests the cli methods to get samples in status-db"""

from cg.store import Store


def test_get_sample_bad_sample(invoke_cli, disk_store: Store):
    """Test to get a sample using a non-existing sample-id """
    # GIVEN an empty database

    # WHEN getting a sample
    db_uri = disk_store.uri
    name = "dummy_name"
    result = invoke_cli(["--database", db_uri, "get", "sample", name])

    # THEN then it should warn about missing sample id instead of getting a sample
    # it will not fail since the API accepts multiple samples
    assert result.exit_code == 0


def test_get_sample_required(invoke_cli, disk_store: Store, helpers):
    """Test to get a sample using only the required argument"""
    # GIVEN a database with a sample
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert disk_store.Sample.query.count() == 1

    # WHEN getting a sample
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id])

    # THEN then it should have been get
    assert result.exit_code == 0
    assert sample_id in result.output


def test_get_samples_required(invoke_cli, disk_store: Store, helpers):
    """Test to get several samples using only the required arguments"""
    # GIVEN a database with two samples
    sample_id1 = "1"
    helpers.add_sample(disk_store, sample_id1)
    sample_id2 = "2"
    helpers.add_sample(disk_store, sample_id2)
    assert disk_store.Sample.query.count() == 2

    # WHEN getting a sample
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id1, sample_id2])

    # THEN then it should have been get
    assert result.exit_code == 0
    assert sample_id1 in result.output
    assert sample_id2 in result.output


def test_get_sample_output(invoke_cli, disk_store: Store, helpers):
    """Test that the output has the data of the sample"""
    # GIVEN a database with a sample with data
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    name = sample.name
    customer_id = sample.customer.internal_id
    application_tag = sample.application_version.application.tag
    state = sample.state
    priority_human = sample.priority_human

    # WHEN getting a sample
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id])

    # THEN then it should have been get
    assert result.exit_code == 0
    assert sample_id in result.output
    assert name in result.output
    assert customer_id in result.output
    assert application_tag in result.output
    assert state in result.output
    assert priority_human in result.output


def test_get_sample_external_false(invoke_cli, disk_store: Store, helpers):
    """Test that the output has the external-value of the sample"""
    # GIVEN a database with a sample with data
    sample = helpers.add_sample(disk_store, is_external=False)
    sample_id = sample.internal_id
    is_external_false = "No"
    is_external_true = "Yes"

    # WHEN getting a sample
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id])

    # THEN then it should have been get
    assert result.exit_code == 0
    assert is_external_false in result.output
    assert is_external_true not in result.output


def test_get_sample_external_true(invoke_cli, disk_store: Store, helpers):
    """Test that the output has the external-value of the sample"""
    # GIVEN a database with a sample with data
    # sample_id = add_sample(disk_store, is_external=True).internal_id
    sample = helpers.add_sample(disk_store, is_external=True)
    sample_id = sample.internal_id
    is_external_false = "No"
    is_external_true = "Yes"

    # WHEN getting a sample
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id])
    # THEN then it should have been get
    assert result.exit_code == 0
    assert is_external_true in result.output
    assert is_external_false not in result.output


def test_get_sample_no_families_without_family(invoke_cli, disk_store: Store, helpers):
    """Test that the --no-families flag works without families"""
    # GIVEN a database with a sample without related samples
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Sample.query.first().links

    # WHEN getting a sample with the --no-families flag
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--no-families"])

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_no_families_with_family(invoke_cli, disk_store: Store, helpers):
    """Test that the --no-families flag doesn't show family info"""
    # GIVEN a database with a sample with related samples
    family = helpers.add_family(disk_store)
    sample = helpers.add_sample(disk_store)
    link = helpers.add_relationship(disk_store, sample=sample, family=family)
    assert link in disk_store.Sample.query.first().links
    sample_id = sample.internal_id

    # WHEN getting a sample with the --no-families flag
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--no-families"])

    # THEN all related families should be listed in the output
    assert result.exit_code == 0
    for link in disk_store.Sample.query.first().links:
        assert link.family.internal_id not in result.output


def test_get_sample_families_without_family(invoke_cli, disk_store: Store, helpers):
    """Test that the --families flag works without families"""
    # GIVEN a database with a sample without related samples
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Sample.query.first().links

    # WHEN getting a sample with the --families flag
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--families"])

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_families_with_family(invoke_cli, disk_store: Store, helpers):
    """Test that the --families flag does show family info"""
    # GIVEN a database with a sample with related samples
    family = helpers.add_family(disk_store)
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    helpers.add_relationship(disk_store, sample=sample, family=family)
    assert disk_store.Sample.query.first().links

    # WHEN getting a sample with the --families flag
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--families"])

    # THEN all related families should be listed in the output
    assert result.exit_code == 0
    for link in disk_store.Sample.query.first().links:
        assert link.family.internal_id in result.output


def test_get_sample_flowcells_without_flowcell(invoke_cli, disk_store: Store, helpers):
    """Test that we can query samples for flowcells even when there are none"""
    # GIVEN a database with a sample without related flowcells
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Flowcell.query.first()

    # WHEN getting a sample with the --flowcells flag
    db_uri = disk_store.uri
    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--flowcells"])

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_flowcells_with_flowcell(invoke_cli, disk_store: Store, helpers):
    """Test that we can query samples for flowcells and that the flowcell name is in the output"""
    # GIVEN a database with a sample and a related flowcell
    flowcell = helpers.add_flowcell(disk_store)
    sample = helpers.add_sample(disk_store, flowcell=flowcell)
    assert flowcell in disk_store.Sample.query.first().flowcells
    sample_id = sample.internal_id

    # WHEN getting a sample with the --flowcells flag
    db_uri = disk_store.uri

    result = invoke_cli(["--database", db_uri, "get", "sample", sample_id, "--flowcells"])

    # THEN the related flowcell should be listed in the output
    assert result.exit_code == 0
    for flowcell in disk_store.Sample.query.first().flowcells:
        assert flowcell.name in result.output



"""CLI support to start microsalt"""

import json
import logging
from pathlib import Path

import click

from cg.apps import hk, lims
from cg.apps.microsalt.fastq import FastqHandler
from cg.cli.workflow.microsalt.store import store as store_cmd
from cg.cli.workflow.microsalt.deliver import deliver as deliver_cmd
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.store import Store
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


@click.group("microsalt", invoke_without_command=True)
@click.option("-d", "--dry-run", is_flag=True, help="print to console")
@click.option("-t", "--ticket", help="include all microbial samples for a ticket")
@click.pass_context
def microsalt(context: click.Context, ticket: str, dry_run: bool):
    """Microbial workflow"""
    context.obj["db"] = Store(context.obj["database"])
    hk_api = hk.HousekeeperAPI(context.obj)
    lims_api = lims.LimsAPI(context.obj)
    analysis_api = MicrosaltAnalysisAPI(
        db=context.obj["db"],
        hk_api=hk_api,
        lims_api=lims_api,
        fastq_handler=FastqHandler(context.obj),
    )
    context.obj["analysis_api"] = analysis_api

    if context.invoked_subcommand:
        return

    if not ticket:
        LOG.error("Please provide a ticket")
        context.abort()

    if not analysis_api.has_flowcells_on_disk(ticket=ticket):
        LOG.warning(
            f"Ticket {ticket} not ready to run, requesting flowcells from long term storage!",
        )
        analysis_api.request_removed_flowcells(ticket=ticket)
        return

    # execute the analysis!
    context.invoke(config_case, ticket=ticket, dry_run=dry_run)
    context.invoke(link, ticket=ticket, dry_run=dry_run)
    context.invoke(run, ticket=ticket, dry_run=dry_run)


@microsalt.command()
@click.option("-d", "--dry-run", is_flag=True, help="only print to console")
@click.option("-t", "--ticket", help="link all microbial samples for a ticket")
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, dry_run: bool, ticket: str, sample_id: str):
    """Link microbial FASTQ files for a SAMPLE_ID"""

    analysis_api = context.obj["analysis_api"]

    if not any([ticket, sample_id]):
        LOG.error("provide ticket and/or sample")
        context.abort()

    if dry_run:
        return

    analysis_api.link_samples(ticket=ticket, sample_id=sample_id)


@microsalt.command("config-case")
@click.option("-d", "--dry-run", is_flag=True, help="print config to console")
@click.option("-t", "--ticket", help="create config-case all microbial samples for an order")
@click.argument("sample_id", required=False)
@click.pass_context
def config_case(context: click.Context, dry_run: bool, ticket: int, sample_id: str):
    """ Create a config file on case level for microSALT """

    if not ticket and not sample_id:
        LOG.error("Provide ticket and/or sample")
        context.abort()

    sample_objs = context.obj["analysis_api"].get_samples(ticket, sample_id)

    if not sample_objs:
        LOG.error("No sample found for that ticket/sample_id")
        context.abort()

    parameters = [
        context.obj["analysis_api"].get_parameters(sample_obj) for sample_obj in sample_objs
    ]

    filename = str(ticket) if ticket else sample_id
    outfilename = (Path(context.obj["microsalt"]["queries_path"]) / filename).with_suffix(".json")
    if dry_run:
        print(json.dumps(parameters, indent=4, sort_keys=True))
        return

    with open(outfilename, "w") as outfile:
        json.dump(parameters, outfile, indent=4, sort_keys=True)


@microsalt.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option(
    "-c",
    "--config-case",
    "config_case_path",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="optionally change the config-case",
)
@click.argument("ticket", type=int)
@click.pass_context
def run(context: click.Context, dry_run: bool, config_case_path: click.Path, ticket: int):
    """ Start microSALT with an order_id """

    microsalt_bin = context.obj["microsalt"]["binary_path"]
    microsalt_env = context.obj["microsalt"]["conda_env"]
    fastq_path = Path(context.obj["microsalt"]["root"]) / "fastq" / str(ticket)

    if config_case_path:
        config_case_path = Path(config_case_path)
    else:
        queries_path = Path(context.obj["microsalt"]["queries_path"])
        config_case_path = (queries_path / str(ticket)).with_suffix(".json")

    process = Process(binary=microsalt_bin, environment=microsalt_env)
    analyse_command = [
        "analyse",
        str(config_case_path.absolute()),
        "--input",
        str(fastq_path.absolute()),
    ]
    process.run_command(parameters=analyse_command, dry_run=dry_run)


microsalt.add_command(config_case)
microsalt.add_command(deliver_cmd)
microsalt.add_command(run)
microsalt.add_command(store_cmd)
